import com.aliasi.cluster.LatentDirichletAllocation;

import com.aliasi.corpus.ObjectHandler;

import com.aliasi.symbol.SymbolTable;

import java.io.File;
import java.io.IOException;

import com.aliasi.util.Files;
import com.aliasi.util.ObjectToCounterMap;
import com.aliasi.util.Strings;

import java.util.List;


public class LdaReportingHandler implements ObjectHandler<LatentDirichletAllocation.GibbsSample> {

    private final long mStartTime;

    LdaReportingHandler(SymbolTable symbolTable) {
        mStartTime = System.currentTimeMillis();
    }

    public void handle(LatentDirichletAllocation.GibbsSample sample) {

        if ((sample.epoch() % 10) == 0) {
            double corpusLog2Prob = sample.corpusLog2Probability();
            System.out.printf("epoch: %5d  token x-entropy: %6.4f\n", 
                              sample.epoch(), 
                              -corpusLog2Prob/sample.numTokens());
        }
    }

    public static void fullReport(LatentDirichletAllocation.GibbsSample sample,
                                  int maxWordsPerTopic,
                                  int maxTopicsPerDoc,
                                  boolean reportTokens,
                                  SymbolTable symbolTable,
				  File outFile) throws IOException {

        System.out.println("\nFull Report");

        int numTopics = sample.numTopics();
        int numWords = sample.numWords();
        int numDocs = sample.numDocuments();
        int numTokens = sample.numTokens();

        System.out.println("epoch=" + sample.epoch());
        System.out.println("numDocs=" + numDocs);
        System.out.println("numTokens=" + numTokens);
        System.out.println("numWords=" + numWords);
        System.out.println("numTopics=" + numTopics);

	
	StringBuilder topicLabels = new StringBuilder();
        for (int topic = 0; topic < numTopics; ++topic) {

	    StringBuilder topicLabel = new StringBuilder();
            int topicCount = sample.topicCount(topic);
            ObjectToCounterMap<Integer> counter = new ObjectToCounterMap<Integer>();
            for (int word = 0; word < numWords; ++word)
                counter.set(Integer.valueOf(word),sample.topicWordCount(topic,word));
            List<Integer> topWords = counter.keysOrderedByCountList();
            System.out.println("\nTOPIC " + topic  + "  (total count=" + topicCount + ")");
            System.out.println("SYMBOL             WORD    COUNT   PROB          Z");
            System.out.println("--------------------------------------------------");
            for (int rank = 0; rank < maxWordsPerTopic && rank < topWords.size(); ++rank) {
                int wordId = topWords.get(rank);
                String word = symbolTable.idToSymbol(wordId);
                int wordCount = sample.wordCount(wordId);
                int topicWordCount = sample.topicWordCount(topic,wordId);
                double topicWordProb = sample.topicWordProb(topic,wordId);
                double z = binomialZ(topicWordCount,
                                     topicCount,
                                     wordCount,
                                     numTokens);
		topicLabel.append(word + " ");
                System.out.printf("%6d  %15s  %7d   %4.3f  %8.1f\n",
                                  wordId,
                                  word,
                                  topicWordCount,
                                  topicWordProb,
                                  z);
            }
	    topicLabels.append(topic + "," + topicLabel + "\n");
        }

	Files.writeStringToFile(topicLabels.toString(),outFile,Strings.UTF8);

        for (int doc = 0; doc < numDocs; ++doc) {
            int docCount = 0;
            for (int topic = 0; topic < numTopics; ++topic)
                docCount += sample.documentTopicCount(doc,topic);
            ObjectToCounterMap<Integer> counter = new ObjectToCounterMap<Integer>();
            for (int topic = 0; topic < numTopics; ++topic)
                counter.set(Integer.valueOf(topic),sample.documentTopicCount(doc,topic));
            List<Integer> topTopics = counter.keysOrderedByCountList();
            System.out.println("\nDOC " + doc);
            System.out.println("TOPIC    COUNT    PROB");
            System.out.println("----------------------");
            for (int rank = 0; rank < topTopics.size() && rank < maxTopicsPerDoc; ++rank) {
                int topic = topTopics.get(rank);
                int docTopicCount = sample.documentTopicCount(doc,topic);
                double docTopicPrior = sample.documentTopicPrior();
                double docTopicProb = (sample.documentTopicCount(doc,topic) + docTopicPrior)
                    / (docCount + numTopics * docTopicPrior);

                System.out.printf("%5d  %7d   %4.3f\n",
                                  topic,
                                  docTopicCount,
                                  docTopicProb);
            }
	    double docTopicPrior = sample.documentTopicPrior();
	    double docTopicProb0 = 1;
	    try {
	       docTopicProb0 = (sample.documentTopicCount(doc,topTopics.get(0)) + docTopicPrior)
	    		    	/ (docCount + numTopics * docTopicPrior);
	  
	   int sym =  sample.word(doc,0);
	   String classTitle = symbolTable.idToSymbol(sym);
	   System.out.printf("## %s %5d %4.3f", classTitle, topTopics.get(0),docTopicProb0);
            System.out.println();
            if (!reportTokens) continue;
            int numDocTokens = sample.documentLength(doc);
            for (int tok = 0; tok < numDocTokens; ++tok) {
                int symbol = sample.word(doc,tok);
                short topic = sample.topicSample(doc,tok);
                String word = symbolTable.idToSymbol(symbol);
                System.out.print(word + "(" + topic + ") ");
            }
            System.out.println();
	    } catch(Exception e) {
	    	System.out.println(e);
	    	System.out.println("document " + doc);
	    }
        }
	    
    }


    static double binomialZ(double wordCountInDoc, double wordsInDoc,
                            double wordCountinCorpus, double wordsInCorpus) {
        double pCorpus = wordCountinCorpus / wordsInCorpus;
        double var = wordsInCorpus * pCorpus * (1 - pCorpus);
        double dev = Math.sqrt(var);
        double expected = wordsInDoc * pCorpus;
	double z = (wordCountInDoc - expected) / dev;
        return z;
    }



}